#pragma once

#include <Eigen/Core>
#include <Eigen/Dense>

#include <memory>
#include <cmath>
#include <limits>
#include <chrono>

struct simple_timer {
    using duration_t = std::chrono::duration<size_t, std::micro>;
    using time_point_t = std::chrono::time_point<std::chrono::steady_clock>;
    time_point_t start = std::chrono::steady_clock::now();
    size_t reset() {
      time_point_t old_start = start;
      start = std::chrono::steady_clock::now();
      return std::chrono::duration_cast<duration_t>(start - old_start).count();
    }
};

// Simple Levenberg-Marquardt solver using dense matrices
template <class T, int N>
struct lm_solver
{
    using matrix = Eigen::Matrix<T, N, N>;
    using vector = Eigen::Matrix<T, N, 1>;

    using calc_func_t = std::function<T(const vector&)>;
    using calc_grad_hessian_t = std::function<void(const vector&, vector&, matrix&)>;

    lm_solver() { }

    lm_solver(calc_func_t calc_func, calc_grad_hessian_t calc_grad_hessian)
      : calc_func(std::move(calc_func)), calc_grad_hessian(std::move(calc_grad_hessian)) { }

    void solve(const vector& x0, T abs_precision = T(0.01), int max_iters = 100, T initial_lambda = T(0.001));

    const vector& get_result() const { return cur_x; }

    size_t get_iteration_count() const { return iter; }

    bool get_verbose() const { return verbose; }
    void set_verbose(bool v) { verbose = v; }

    T get_start_S() const { return start_S; }
    T get_end_S() const { return end_S; }

  private:
    calc_func_t calc_func;
    calc_grad_hessian_t calc_grad_hessian;

    bool verbose = true;
    int iter = 0;
    vector cur_x;
    T start_S {0}, end_S {0};
};

template <class T, int N>
void lm_solver<T, N>::solve(const vector& x0, T abs_precision, int max_iters, T initial_lambda) {
    if ( N == Eigen::Dynamic && x0.size() == 0 ) {
        fprintf(stderr, "lm_solver: input vector has zero length\n");
        return;
    }

    simple_timer timer;

    cur_x = x0;

    T lambda = initial_lambda;

    vector try_x, grad;

    start_S = calc_func(cur_x);
    end_S = start_S;

    matrix H;
    vector diagonal;
    if ( N == Eigen::Dynamic ) {
        H.resize(x0.size(), x0.size());
        diagonal.resize(x0.size());
        grad.resize(x0.size());
    }

    int solve_count = 0;
    int small_count = 0;
    bool done = false;

    int n = (N == Eigen::Dynamic ? x0.size() : N);

    for ( iter = 0; iter < max_iters && !done; ) {
        calc_grad_hessian(cur_x, grad, H);

        diagonal = H.diagonal();

        for ( int i = 0; i < n; i++ ) {
            if ( diagonal[i] <= 0 )
                fprintf(stderr, "lm_solver: CAUTION: zero or negative diagonal entry at %d\n", i);
        }

        int inner = 0;
        const int INNER_LOOPS = 20;
        for ( ; inner < INNER_LOOPS && iter < max_iters; inner++ ) {
            ++iter;

            matrix H_damped = H;
            H_damped.diagonal() = (1 + lambda) * diagonal;

            Eigen::LDLT<matrix> solver(H_damped);

            if ( solver.info() != Eigen::Success ) {
                fprintf(stderr, "lm_solver: LDLT solver failed, falling back to FullPivHouseholderQR\n");
                Eigen::FullPivHouseholderQR<matrix> solver(H_damped);
                try_x = cur_x - solver.solve(grad);
            } else {
                try_x = cur_x - solver.solve(grad);
            }

            solve_count++;

            T new_S = calc_func(try_x);

            if ( verbose )
                fprintf(stderr, "%d/%d: %g->%g / %g, lambda=%g (step=%zu)\n", iter, inner, end_S, new_S, end_S - new_S, lambda, timer.reset());

            if ( std::isfinite(new_S) && new_S < end_S ) {
                if ( lambda > T(1e-12) )
                    lambda *= T(0.2);
                lambda = std::max(lambda, std::numeric_limits<T>::epsilon());
                cur_x = try_x;

                if ( iter >= 5 && (end_S - new_S) < abs_precision ) {
                    small_count++;
                    if ( small_count >= 3 )
                        done = true;
                } else {
                    small_count = 0;
                }
                end_S = new_S;
                break;
            }

            lambda *= 10;

            if ( lambda > T(1e10) ) {
                inner = INNER_LOOPS;
                break;
            }
        }
        if ( inner == INNER_LOOPS )
            break;
    }
}

