
#include <Eigen/Core>
#include <vector>

#include "lm_solver.hpp"

struct LinearModel {
    using vector = Eigen::Vector2d;
    using matrix = Eigen::Matrix2d;
    double sigma = 1, gamma = 2.0 * std::sqrt(25) / M_PI;
    double a = 0, b = 1;
    struct sensor_reading { double x, y; };
    std::vector<sensor_reading> sensor_data;
    void calc_S_gradient_hessian(const vector& M, double& S, vector& G, matrix& H) {
        double a = M(0); double b = M(1);
        S = 0; G = G.Zero(); H = H.Zero();
        for ( auto& sensor : sensor_data ) {
            // Auto-generated code
            double x0 = 1.0/sigma;
            double x1 = a*sensor.x + b - sensor.y;
            double x2 = x0/(1 + pow(x1, 2)/(pow(gamma, 2)*pow(sigma, 2)));
            double r = gamma*atan(x0*x1/gamma);
            double drda = sensor.x*x2;
            double drdb = x2;
            S += r*r;
            G(0) += 2*r*drda;
            G(1) += 2*r*drdb;
            H(0,0) += 2*drda*drda;
            H(0,1) += 2*drda*drdb;
            H(1,1) += 2*drdb*drdb;
        }
        H(1,0) = H(0,1); // H is symmetric
    }
    // ...

    double calc_S(const vector& M)
    {
        const double a = M(0), b = M(1);
        double S = 0;
        for ( auto& sensor : sensor_data ) {
            double x0 = 1.0/sigma;
            double x1 = a*sensor.x + b - sensor.y;
            double r = gamma*atan(x0*x1/gamma);
            S += r*r;
        }
        return S;
    }

    // Fit using simple linear regression. Good for initial values
    void fit_linear_regression()
    {
      double sum_x = 0, sum_x2 = 0, sum_y = 0, sum_xy = 0;
      for ( auto& e : sensor_data ) {
          double x = e.x, y = e.y;
          sum_x += x;
          sum_x2 += x * x;
          sum_y += y;
          sum_xy += x * y;
      }
      double n = sensor_data.size();
      a = (n * sum_xy - sum_x * sum_y) / (n * sum_x2 - sum_x * sum_x);
      b = (sum_y - a * sum_x ) / n;
    }

    // Nonlinear minimization with suppressed errors
    void fit(int iters = 10, double lambda = 0.001)
    {
      lm_solver<double, 2> opt(
          [this](const vector& M) { return calc_S(M); },
          [this](const vector& M, vector& G, matrix& H) { double S; calc_S_gradient_hessian(M, S, G, H); }
      );
      opt.solve({a, b}, 0.01, iters, lambda);
      auto& res = opt.get_result();
      a = res(0);
      b = res(1);
    }
};

